/*
This file used to be part of mfaktc.
Copyright (C) 2009, 2010, 2011  Oliver Weihe (o.weihe@t-online.de)
This file has been written by Luigi Morelli (L.Morelli@mclink.it) *1

mfaktc is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mfaktc is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mfaktc.  If not, see <http://www.gnu.org/licenses/>.
*/



/*
*1 Luigi initially wrote the two functions get_next_assignment() and
clear_assignment() after we (Luigi and myself (Oliver)) have discussed the
interface. Luigi was so nice to write those functions so I had time to focus
on other parts, this made early (mfaktc 0.07) worktodo parsing posible.
For mfakc 0.15 I've completly rewritten those two functions. The rewritten
functions should be more robust against malformed input. Grab the sources of
mfaktc 0.07-0.14 to see Luigis code.
*/

//Christenson, 2011, needed to expand the functionality, and strongly
// disliked not running everything through the same parse function.
//  after a first re-write with a do-everything parse function that was
//  too complicated for its own good, a second re-write centers around
//  a new function, parse_line, that takes in a line of worktodo and returns
//  a structured result.
// the value of this is that it can be re-targeted to some other forms of
// numbers besides Mersenne exponents.....

/* Bill stole it and modified it for CUDALucas' purposes... :) */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#define MAX_LINE_LENGTH 256

#ifndef _MSC_VER /* See next comment */
  #define _fopen fopen
  #define strcopy strncpy
  #define _sprintf sprintf
  #include <unistd.h>
  #include <sched.h>
  #define MODE S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH
  static void Sleep(unsigned int ms)
  {
    struct timespec ts;
    ts.tv_sec  = (time_t) ms/1000;
    ts.tv_nsec = (ms % 1000) * 1000000;
    nanosleep(&ts, NULL);
  }
#else
  #include <winsock2.h>
  #include <io.h>
  #undef open
  #undef close
  #define open _open
  #define close _close
  #define sched_yield SwitchToThread
  #define MODE _S_IREAD | _S_IWRITE
  #define strncasecmp _strnicmp

  /* Everything from here to the next include is to make MSVS happy. */
  #define sscanf sscanf_s /* This only works for scanning numbers, or strings with a defined length (e.g. "%131s") */
  void strcopy(char* dest, char* src, size_t n)
  {
    strncpy_s(dest, MAX_LINE_LENGTH, src, n);
  }
  FILE* _fopen(const char* path, const char* mode)
  {
    FILE* stream;
    errno_t err = fopen_s(&stream, path, mode);
    if(err) return NULL;
    else return stream;
  }
  void _sprintf(char* buf, char* frmt, const char* string)
  { // only used in filelocking code
	  sprintf_s(buf, 251, frmt, string);
  }

  int gettimeofday(struct timeval *tv, struct timezone *unused)
  /*
  This is based on a code sniplet from Kevin (kjaget on www.mersenneforum.org)

  This doesn't act like a real gettimeofday(). It has a wrong offset but this is
  OK since CUDALucas only uses this to measure the time difference between two calls
  of gettimeofday().
  */
  {
    static LARGE_INTEGER frequency;
    static int frequency_flag = 0;

    if(!frequency_flag)
    {
      QueryPerformanceFrequency(&frequency);
      frequency_flag = 1;
    }

    if(tv)
    {
      LARGE_INTEGER counter;
      QueryPerformanceCounter(&counter);
      tv->tv_sec =  (long) (counter.QuadPart / frequency.QuadPart);
      tv->tv_usec = (long)((counter.QuadPart % frequency.QuadPart) / ((double)frequency.QuadPart / 1000000.0));
    }
    return 0;
  }
#endif

/***********************************************************************************************************/

int file_exists(char* name) {
  if(name == NULL || name[0] == 0)
    return 0;

  FILE* stream = _fopen(name, "r");
  if(stream != NULL)
  {
    fclose(stream);
    return 1;
  }

  return 0;
}

int isprime(unsigned int n)
{
  if(n <= 1) return 0; // 0, 1
  if(n <= 3) return 1; // 2, 3
  if(n%2==0 
  || n%3==0) return 0; // do this, so that we can limit to i=+/-1 mod 6 in the next loop
  if(n < 25) return 1; // because, why not

  for(unsigned int i=5; (i <= 65521) && (i*i <= n); i+=6) // loop thru i = -1 mod 6. 65521 is the biggest 16-bit prime
  {
    if(n%i==0 || n%(i+2)==0) return 0; // check the pair -1 mod 6 and +1 mod 6
  }
  return 1;
}

int valid_assignment(int exp, int fftlen)
{
  if(exp < 150000) // arbitrary
  {
    fprintf(stderr, "Warning: exponents < 150000 are not supported!\n");
    return 0;
  }
  if(exp > 100000000) // arbitrary
  {
    fprintf(stderr, "Warning: exponents > 100M are not supported!\n");
    return 0;
  }
  if(fftlen)
  {
  if(fftlen % 1024)
  {
    fprintf(stderr, "Warning: FFT length %d is invalid, it must be a multiple of 1024. See cudaWagstaff.ini for more details about good lengths.\n",
            fftlen);
    return 0;
  }
  if(fftlen < 8*1024 || fftlen > 8*1024*1024) // support 8K till 8M for now. consistent with exponent range.
  {
    fprintf(stderr, "Warning: FFT length %d is not supported currently, it must be between 8K and 8M.\n",
            fftlen);
    return 0;
  }
  }
// for now this is fine, but we should be able to support exponents of form 2^k*q without much issue
// (2^(2*q)+1)/5 is not interesting due to Aurifeuillean factorization
// (2^(4*q)+1)/17 -- http://oeis.org/A057182
// (2^(8*q)+1)/257 is ok
// etc
  if(!isprime(exp)) 
  {
    fprintf(stderr, "Warning: exponent is not prime!\n");
    return -1;
  }

  return 1;
}

enum ASSIGNMENT_ERRORS
{	NEVER_ASSIGNED=-1,
	OK=0,
	CANT_OPEN_FILE=1,
	VALID_ASSIGNMENT_NOT_FOUND=2,
	CANT_OPEN_WORKFILE=3,
	CANT_OPEN_TEMPFILE=4,
	ASSIGNMENT_NOT_FOUND=5,
	CANT_RENAME =6
};

typedef char LINE_BUFFER[MAX_LINE_LENGTH];

enum PARSE_WARNINGS
{
  NO_WARNING=0,
  END_OF_FILE,
  LONG_LINE,
  NO_TEST_EQUAL,
  INVALID_FORMAT,
  INVALID_DATA,
  BLANK_LINE,
  NONBLANK_LINE
};

struct ASSIGNMENT
{
	int exponent;
	int fft_length;
};

enum PARSE_WARNINGS parse_worktodo_line(FILE *f_in, struct ASSIGNMENT *assignment, LINE_BUFFER *linecopy)
/*
output
  assignment: structure of line, with any assignment if found
  linecopy: a copy of the last read line
*/
{
  char line[MAX_LINE_LENGTH], *ptr;

  unsigned long proposed_exponent = 0;
  unsigned long proposed_fftlen = 0;
  int digit_count;

  assignment->fft_length = 0;
  assignment->exponent = 0;

  if(NULL==fgets(line, MAX_LINE_LENGTH, f_in))
  {
    return END_OF_FILE;
  }

  // if a copy is requested, return what was read.
  if (linecopy != NULL)	{ 
    strcopy(*linecopy, line, MAX_LINE_LENGTH);
    ptr = strchr(*linecopy, '\n');
    if(ptr != NULL)
      *ptr = '\0';
  }

  ptr=line;
  while (*ptr && isspace(*ptr))	ptr++; // skip blanks

  if (!*ptr)	// eol
    return BLANK_LINE;

  // comments. lazy. 
  if( ('\\'== *ptr)) return NONBLANK_LINE;
  if( ('/' == *ptr)) return NONBLANK_LINE;
  if( ('#' == *ptr)) return NONBLANK_LINE;

  // does the line start with "PRP=" ?
  if ( (strncasecmp("PRP=", ptr, 4) != 0) )
    return NO_TEST_EQUAL;

  ptr += 4;

  while (*ptr && isspace(*ptr))	ptr++; // skip blanks
    

  // Business end of parsing. We support PRP=exponent and PRP=exponent,FFT
  proposed_exponent = 0; 

  digit_count = 0;
  while (*ptr && isdigit(*ptr))	// grab digits
  {
    proposed_exponent = proposed_exponent*10 + (*ptr - '0');
    ptr++;
    digit_count++;
  }
  if(digit_count < 4 || digit_count > 9) // sanity check
    return INVALID_DATA;
  
  while (*ptr && isspace(*ptr))	ptr++; // skip blanks

  if (*ptr == ',')  // fft length specified, grab it
  {
      ptr++;
      while (*ptr && isspace(*ptr))	ptr++; // skip blanks

	  proposed_fftlen = 0;
      digit_count = 0;

	  while (*ptr && isdigit(*ptr))	// grab digits
	  {
		proposed_fftlen = proposed_fftlen*10 + (*ptr - '0');
		ptr++;
        digit_count++;
	  }
	  if(digit_count == 0 || digit_count > 7) // sanity check
        return INVALID_DATA;

      if(*ptr == 'k' || *ptr == 'K') // need to multiply by 1024
      {
        if(digit_count > 5) return INVALID_DATA; // sanity check
        proposed_fftlen *= 1024;
        ptr++;
      }
      else if(*ptr == 'm' || *ptr == 'M') // need to multiply by 1024*1024
      {
        if(digit_count > 2) return INVALID_DATA; // sanity check
        proposed_fftlen *= 1024*1024;
        ptr++;
      }
      while (*ptr && isspace(*ptr))	ptr++; // skip blanks
  }
  
  // See if anything unprocessable is there at the end.
  if( ('\\'== *ptr) // comment
    ||('/'== *ptr)  // comment
    ||('#'== *ptr)  // comment
    ||('\n'== *ptr) // new line
    ||('\0'== *ptr) // eol
    )
  {
    
  }
  else // Oops. Something bad left behind
    return INVALID_FORMAT;

  if(proposed_fftlen > proposed_exponent) // sanity check
    return INVALID_DATA;

  assignment->exponent = (int) proposed_exponent;
  assignment->fft_length = (int) proposed_fftlen;

  return NO_WARNING;
}

/**************************************************************
 * Function name : get_next_assignment                        *
 *                                                            *
 *     OUTPUT :                                               *
 *                                                            *
 *     0 - OK                                                 *
 *     1 - get_next_assignment : cannot open file			  *
 *     2 - get_next_assignment : no valid assignment found    *
 **************************************************************/
enum ASSIGNMENT_ERRORS get_next_assignment(char *filename, int *exponent, int* fft_length)
{
  struct ASSIGNMENT assignment;

  FILE *f_in;
  enum PARSE_WARNINGS value;

  LINE_BUFFER line;
  unsigned int linecount = 0;

  assignment.fft_length = 0;
  f_in = _fopen(filename, "r");
  if(NULL == f_in)
  {
    fprintf(stderr, "Can't open workfile %s\n\n", filename);
    return CANT_OPEN_FILE;	// nothing to open...
  }

  for(;;)
  {
    linecount++;

    value = parse_worktodo_line(f_in,&assignment,&line);

    if ((BLANK_LINE == value) || (NONBLANK_LINE == value))
      continue;
    if (NO_WARNING == value)
    {
      if (valid_assignment(assignment. exponent, assignment. fft_length) > 0)
        break;
      value = INVALID_DATA;
    }

    if (END_OF_FILE == value)
      break;
    fprintf(stderr, "Warning: ignoring line %u: \"%s\" in \"%s\". Reason: ", linecount, line, filename);
    switch(value)
    {
      case LONG_LINE:           printf("line is too long.\n"); break;
      case NO_TEST_EQUAL:       printf("doesn't begin with PRP=.\n");break;
      case INVALID_FORMAT:      printf("invalid format.\n");break;
      case INVALID_DATA:        printf("invalid data.\n");break;
      default:                  printf("unknown error.\n"); break;
    }

  }

  fclose(f_in);

  if (NO_WARNING == value)
  {
    *exponent = assignment.exponent;
    if(assignment.fft_length > 0) *fft_length = assignment.fft_length;

    return OK;
  }

  fprintf(stderr, "No valid assignment found.\n\n");
  return VALID_ASSIGNMENT_NOT_FOUND;
}

enum ASSIGNMENT_ERRORS clear_assignment(char *filename, int exponent)
{
  int found = 0;
  FILE *f_in, *f_out;
  LINE_BUFFER line;	// line buffer

  enum PARSE_WARNINGS value;

  struct ASSIGNMENT assignment;

  f_in = _fopen(filename, "r");
  if (NULL == f_in) {
    fprintf(stderr, "Can't open workfile %s\n\n", filename);
    return CANT_OPEN_WORKFILE;
  }

  f_out = _fopen("__worktodo__.tmp", "w");
  if (NULL == f_out)
  {
    fclose(f_in);
    fprintf(stderr, "Can't open tmp workfile\n\n");
    return CANT_OPEN_TEMPFILE;
  }

  found = 0;
  while (END_OF_FILE != (value = parse_worktodo_line(f_in,&assignment,&line)) )
  {
    if(NO_WARNING == value && (exponent == assignment.exponent) ) // current assignment line?
    {
      found = 1; // set the flag, but don't copy the line to tmp
      // WARNING: if, for some reason, same assignment is there multiple times, it will get rid of all of them.
      // Fix is simple, but I don't want to overcomplicate
    }
    else
    {
      fprintf(f_out,"%s\n",line); // copy the line to tmp, 'cos this ain't the assigment we're looking for
    }
  }

  fclose(f_in);
  fclose(f_out);

  if (!found) {
    fprintf(stderr, "Assignment W%d completed but not found in workfile\n\n", exponent);
    return ASSIGNMENT_NOT_FOUND;
  }
  if(remove(filename) != 0) {
    fprintf(stderr, "Cannot remove old workfile\n\n");
    return CANT_RENAME;
  }
  if(rename("__worktodo__.tmp", filename) != 0) {
    fprintf(stderr, "Cannot move tmp workfile to regular workfile\n\n");
    return CANT_RENAME;
  }
  return OK;
}

/*----------------------------------------------------------------------------------------------*/
/* These functions are slightly more modified from Oliver's equivalents. */

int IniGetInt(char *inifile, char *name, int *value, int dflt)
{
  FILE *in;
  char buf[MAX_LINE_LENGTH];
  int found=0;
  in=_fopen(inifile,"r");
  if(!in) goto error;
  while(fgets(buf,MAX_LINE_LENGTH-1,in) && !found)
  {
    if(!strncmp(buf,name,strlen(name)) && buf[strlen(name)]=='=')
    {
      if(sscanf(&(buf[strlen(name)+1]),"%d",value)==1)found=1;
    }
  }
  fclose(in);
  if(found) return 1;
  error: *value = dflt; return 0;
}

int IniGetInt2(char *inifile, char *name, int *value0, int *value1, int dflt)
{
  FILE *in;
  char buf[MAX_LINE_LENGTH];
  int found=0;
  in=_fopen(inifile,"r");
  if(!in) goto error;
  while(fgets(buf,MAX_LINE_LENGTH-1,in) && !found)
  {
    if(!strncmp(buf,name,strlen(name)) && buf[strlen(name)]=='=')
    {
      if(sscanf(&(buf[strlen(name)+1]),"%d %d",value0, value1)==2)found=1;
    }
  }
  fclose(in);
  if(found) return 1;
  error: *value0 = dflt; return 0;
}

int IniGetInt3(char *inifile, char *name, int *value0, int *value1, int *value2, int dflt)
{
  FILE *in;
  char buf[MAX_LINE_LENGTH];
  int found=0;
  in=_fopen(inifile,"r");
  if(!in) goto error;
  while(fgets(buf,MAX_LINE_LENGTH-1,in) && !found)
  {
    if(!strncmp(buf,name,strlen(name)) && buf[strlen(name)]=='=')
    {
      if(sscanf(&(buf[strlen(name)+1]),"%d %d %d",value0, value1, value2)==3)found=1;
    }
  }
  fclose(in);
  if(found) return 1;
  error: *value0 = dflt; return 0;
}

int IniGetStr(char *inifile, char *name, char *string, char* dflt)
{
  FILE *in;
  char buf[MAX_LINE_LENGTH];
  int found=0;
  int idx = strlen(name);
  int len = MAX_LINE_LENGTH;
  in=_fopen(inifile,"r");
  if(!in)
    goto error;
  while(fgets(buf,MAX_LINE_LENGTH - 1,in) && !found)
  {
    if(!strncmp(buf,name,idx) && buf[idx]=='=')
    {
      found = strlen(buf + idx + 1);
      found = (len > found ? found : len) - 1;
      if (found)
      {
        strncpy(string, buf + idx + 1, found);
        if(string[found - 1] == '\r') found--; //remove '\r' from string, this happens when reading a DOS/Windows formatted file on Linux
      }
      string[found] = '\0';
    }
  }
  fclose(in);
  if(found)return 1;
  error: strcopy(string, dflt, MAX_LINE_LENGTH); return 0;
}

/*****************************************************************************/
/*      mfakto's file locking code                                           */

#define MAX_LOCKED_FILES 3

typedef struct _lockinfo
{
  int       lockfd;
  FILE *    open_file;
  char      lock_filename[256];
} lockinfo;

static unsigned int num_locked_files = 0;
static lockinfo     locked_files[MAX_LOCKED_FILES];

FILE *fopen_and_lock(const char *path, const char *mode)
{
  unsigned int i;
  int lockfd;
  FILE *f;
  #ifdef EBUG
  printf("\nlock() called on %s\n", path);
  #endif

  if (strlen(path) > 250)
  {
    fprintf(stderr, "Cannot open %.250s: Name too long.\n", path);
    return NULL;
  }

  if (num_locked_files >= MAX_LOCKED_FILES)
  {
    fprintf(stderr, "Cannot open %.250s: Too many locked files.\n", path);
    return NULL;
  }

  _sprintf(locked_files[num_locked_files].lock_filename, "%.250s.lck", path);

  for(i=0;;)
  {
    if ((lockfd = open(locked_files[num_locked_files].lock_filename, O_EXCL | O_CREAT, MODE)) < 0)
    {
      if (errno == EEXIST)
      {
        if (i==0) fprintf(stderr, "%.250s is locked, waiting ...\n", path);
        if (i<1000) i++; // slowly increase sleep time up to 1 sec
        Sleep(i);
        continue;
      }
      else
      {
        perror("Cannot open lockfile");
        break;
      }
    }
    break;
  }

  locked_files[num_locked_files].lockfd = lockfd;

  if (lockfd > 0 && i > 0)
  {
    printf("Locked %.250s\n", path);
  }

  f=fopen(path, mode);
  if (f)
  {
    locked_files[num_locked_files++].open_file = f;
  }
  else
  {
    if (close(locked_files[num_locked_files].lockfd) != 0) perror("Failed to close lockfile");
    if (remove(locked_files[num_locked_files].lock_filename)!= 0) perror("Failed to delete lockfile");
  }
  #ifdef EBUG
  printf("successfully locked %s\n", path);
  #endif
  #ifdef TEST
  while(1);
  #endif

  return f;
}

int unlock_and_fclose(FILE *f)
{
  unsigned int i, j;
  int ret;
  #ifdef EBUG
  printf("unlock() called\n");
  #endif

  if (f == NULL) return -1;

  for (i=0; i<num_locked_files; i++)
  {
    if (locked_files[i].open_file == f)
    {
      ret = fclose(f);
      f = NULL;
      if (close(locked_files[i].lockfd) != 0) perror("Failed to close lockfile");
      if (remove(locked_files[i].lock_filename)!= 0) perror("Failed to delete lockfile");
      for (j=i+1; j<num_locked_files; j++)
      {
        locked_files[j-1].lockfd = locked_files[j].lockfd;
        locked_files[j-1].open_file = locked_files[j].open_file;
        strcpy(locked_files[j-1].lock_filename, locked_files[j].lock_filename);
      }
      num_locked_files--;
      break;
    }
  }
  if (f)
  {
    fprintf(stderr, "File was not locked!\n");
    ret = fclose(f);
  }
  #ifdef EBUG
  printf("successfully unlocked\n");
  #endif
  return ret;
}
